unit pdf_edit_text;

interface

uses Classes, Math, SysUtils, dynapdf;

{
   The class CPDFEditText demonstrates how texts can be found and replaced in a PDF file.
   The algorithm handles rotated text as well as text lines which consist of multiple
   text records correctly.

   This algorithm assumes that the text is ordered from left to right and from top
   to bottom. If this is not the case the result is maybe incorrect or the search
   text cannot be found.

   The remaing text in a line is realigned to avoid overlaping text or holes
   between text redords. This realignment is not a real text re-flow because it is
   restricted to the current text line. No line break occurs if the new text is too
   large to fit into the page or text block. It is possible to develop a real text
   re-flow but such an algorithm is much more complex than this one.

   A page must always be parsed twice, the first time to find the text records which
   make up the search string, and a second time to delete or replace the text.

   ----------------------------------------------------------------------------------

   A note about underlined text:
   The underline of a text is not a font attribute. It is a separate vector graphic
   which occurs outside of the text object, e.g. as a rectanlge or a line drawn with
   a moveto/lineto pair. DynaPDF contains no algorithms to identify such underlines so
   that they can be deleted with the text. If you delete an underlined text the
   underline will be left unchanged and only the text is deleted!
}

{
   Calculation of the space width:

   Space characters are often not stored in PDF files, instead, kerning space
   is applied which produces the same visual appearance. Due to the missing space
   characters we need an algorithm that is able to detect word boundaries.

   The algorithm here compares just the distance between two text records and the
   the displacement in kerning records. If it is larger than the half space width
   we assume that a space character was emulated at this position. Note that the
   displacement in a kerning record is measured in text space while the distance
   between two text records is measured in user space!

   The full space width is often too large to determine whether normal kerning
   space was applied or whether a space character must be added at a given position.
   The half space width procudes better results in most cases especially if the
   document uses condensed fonts. The fonts of documents which emulate space
   characters contain often no space character. In this case a standard space width
   is set because the real value is not known.
}

type TTextRec = record
   TmplHandle: Integer;
   First:      Integer;
   Last:       Integer;
   KernRecord: Integer;
   StrPos:     Integer;
   NewLine:    Boolean;
end;
type TTextRecPtr = ^TTextRec;

// List to store template handles
type CIntList = class(TObject)
  private
   m_Capacity: Integer;
   m_Count:    Integer;
   m_Items:    PInteger;
  public
   destructor  Destroy(); override;
   procedure   Add(Value: Integer); 
   procedure   Clear();
   property    Count: Integer read m_Count;
   function    GetItem(Index: Cardinal): Integer;
end;

// List to store text records
type CRecordList = class(TObject)
  private
   m_Capacity: Integer;
   m_Count:    Integer;
   m_Items:    TTextRecPtr;
  public
   destructor  Destroy(); override;
   procedure   Add(TmplHandle, First, Last, KernRecord, StrPos: Integer; NewLine: Boolean); 
   procedure   Clear();
   property    Count: Integer read m_Count;
   function    GetItem(Index: Cardinal): TTextRecPtr;
end;

type CPDFEditText = class(TObject)
 public
   constructor Create(const PDFInst: TPDF);
   destructor  Destroy(); override;
   function    FindText(const Text: WideString): Integer;
   procedure   ReplacePattern(const NewText: WideString);
  protected
   m_Alpha:         Integer;
   m_CurrTmpl:      Integer;
   m_First:         Integer;
   m_HaveMore:      Boolean;
   m_KernRecord:    Integer;
   m_LastFont:      Pointer;
   m_LastX:         Double;
   m_LastY:         Double;
   m_Matrix1:       TCTM;
   m_Matrix2:       TCTM;
   m_Matrix3:       TCTM;
   m_Matrix4:       TCTM;
   m_NewLine:       Boolean;
   m_PDFInst:       TPDF;
   m_RecordNumber:  Integer;
   m_Records:       CRecordList;
   m_SearchPos:     PWideChar;
   m_SearchText:    PWideChar;
   m_SearchTextLen: Integer;
   m_Stack:         TPDFStack;
   m_StrPos:        Integer;
   m_Templates:     CIntList;

   procedure AddKernSpace(Advance, SpaceWidth: Single; var Matrix: TCTM);
   procedure AddKernSpaceEx(Advance, SpaceWidth: Single; var Matrix: TCTM);
   procedure AddRecord(KernRecord, StrPos: Integer);
   procedure AddSpace(x, y, SpaceWidth: Double); overload;
   procedure AddSpace(DistX, SpaceWidth: Double; var Matrix: TCTM); overload;
   function  BildFamilyNameAndStyle(FontName: PChar; Len: Integer; FamilyName: PChar; var Style: TFStyle): Boolean;
   procedure CalcAlpha(var M: TCTM);
   procedure CalcDistance(var M1, M2, M3: TCTM; var DistX, DistY: Double; x, y: Double); overload;
   function  CalcDistance(x1, y1, x2, y2: Double): Double; overload;
   procedure CalcStrPos(var M: TCTM; var x, y: Double);
   procedure FindAll();
   function  FindChar(Text: PChar; AChar: Char): Integer;
   function  FindEndPattern(const Text: PWideChar; Len: Integer): Integer;
   procedure FindPattern();
   function  GetScaleFactor(var M: TCTM): Double;
   procedure Invert(var M: TCTM);
   function  MulMatrix(var M1, M2: TCTM): TCTM;
   procedure ParseTemplates();
   procedure SetFont();
   procedure Transform(var M: TCTM; var x, y: Double);
   function  WriteLine(Index: Integer; TextRecord: TTextRecPtr; NewText: PWideChar; Len: Integer): Integer;
   procedure WriteRecord(var M: TCTM; TextRecord: TTextRecPtr; Source: TTextRecordWPtr; NewText: PWideChar; Len: Integer);
   function  WriteRemaining(Index: Integer; TextRecord: TTextRecPtr; NewText: PWideChar; Len: Integer): Integer;
   procedure WriteText(var M: TCTM; Text: PWideChar; Len: Integer);
end;

implementation

{ CIntList }

procedure CIntList.Add(Value: Integer);
var item: PInteger;
begin
	if m_Count + 1 > m_Capacity then begin
		ReallocMem(m_Items, m_Capacity + 4196);
		Inc(m_Capacity, 4196);
	end;
   item := m_Items;
   Inc(item, m_Count);
   item^ := Value;
	Inc(m_Count);
end;

procedure CIntList.Clear;
begin
   m_Count := 0;
end;

destructor CIntList.Destroy;
begin
   if m_Items <> nil then FreeMem(m_Items);
   inherited;
end;

function CIntList.GetItem(Index: Cardinal): Integer;
var item: PInteger;
begin
   item := m_Items;
   Inc(item, Index);
   Result := item^;
end;

{ CRecordList }

destructor CRecordList.Destroy;
begin
   if m_Items <> nil then FreeMem(m_Items);
   inherited;
end;

procedure CRecordList.Add(TmplHandle, First, Last, KernRecord, StrPos: Integer; NewLine: Boolean);
var item: TTextRecPtr;
begin
	if m_Count + 1 > m_Capacity then begin
		ReallocMem(m_Items, m_Capacity + 4196);
		Inc(m_Capacity, 4196);
	end;
   item := m_Items;
   Inc(item, m_Count);
	item.TmplHandle := TmplHandle;
	item.First      := First;
	item.Last       := Last;
	item.KernRecord := KernRecord;
	item.StrPos     := StrPos;
	item.NewLine    := NewLine;
	Inc(m_Count);
end;

procedure CRecordList.Clear;
begin
   m_Count := 0;
end;

function CRecordList.GetItem(Index: Cardinal): TTextRecPtr;
begin
   Result := m_Items;
   Inc(Result, Index);
end;

{ CPDFEditText }

const MAX_LINE_ERROR: Double = 2.0;

constructor CPDFEditText.Create(const PDFInst: TPDF);
begin
   m_First     := -1;
	m_Matrix4.a := 1.0;
	m_Matrix4.d := 1.0;
   m_NewLine   := true;
	m_PDFInst   := PDFInst;
	m_Records   := CRecordList.Create;
	m_Templates := CIntList.Create;
end;

destructor CPDFEditText.Destroy;
begin
  if m_Records   <> nil then m_Records.Free;
  if m_Templates <> nil then m_Templates.Free;
  inherited;
end;

procedure CPDFEditText.AddKernSpace(Advance, SpaceWidth: Single; var Matrix: TCTM);
begin
   if Advance < SpaceWidth then begin
      if m_SearchPos^ = #32 then begin
         Inc(m_SearchPos);
         if m_SearchPos^ = #0 then m_SearchPos := m_SearchText;
      end else begin
         m_Matrix4.x := -Advance;
         Matrix      := MulMatrix(Matrix, m_Matrix4);
      end;
   end else begin
      m_Matrix4.x := -Advance;
      Matrix      := MulMatrix(Matrix, m_Matrix4);
   end;
end;

procedure CPDFEditText.AddKernSpaceEx(Advance, SpaceWidth: Single; var Matrix: TCTM);
begin
   if Advance < SpaceWidth then begin
      if m_SearchPos^ = #32 then begin
         Inc(m_SearchPos);
      end else begin
         m_Matrix4.x := -Advance;
         Matrix      := MulMatrix(Matrix, m_Matrix4);
      end;
   end else begin
      m_Matrix4.x := -Advance;
      Matrix      := MulMatrix(Matrix, m_Matrix4);
   end;
end;

procedure CPDFEditText.AddRecord(KernRecord, StrPos: Integer);
begin
   m_Records.Add(m_CurrTmpl, m_First, m_RecordNumber, KernRecord, StrPos, m_NewLine);
   m_KernRecord := -1;
   m_NewLine    := false;
   m_SearchPos  := m_SearchText;
   m_StrPos     := -1;
end;

procedure CPDFEditText.AddSpace(x, y, SpaceWidth: Double);
var distX: Double;
begin
   if m_Alpha = 0 then 
      distX := x - m_LastX
   else
      distX := CalcDistance(m_LastX, m_LastY, x, y);
   if distX > SpaceWidth then begin
      if m_SearchPos^ = #32 then Inc(m_SearchPos);
   end;
end;

procedure CPDFEditText.AddSpace(DistX, SpaceWidth: Double; var Matrix: TCTM);
begin
   if DistX > SpaceWidth then begin
      if m_SearchPos^ = #32 then
         Inc(m_SearchPos)
      else begin
         m_Matrix4.x := DistX;
         Matrix      := MulMatrix(Matrix, m_Matrix4);
      end;
   end else begin
      m_Matrix4.x := DistX;
      Matrix      := MulMatrix(Matrix, m_Matrix4);
   end;
end;

function CPDFEditText.BildFamilyNameAndStyle(FontName: PChar; Len: Integer; FamilyName: PChar; var Style: TFStyle): Boolean;
var i, p: Integer; fName: PChar;
begin
   if (Len > 7) and (FontName[6] = '+') then begin
      Inc(FontName, 7);
      Dec(Len, 7);
      if Len > 64 then Len := 64;
   end;
   p := FindChar(FontName, ',');
   if p > -1 then begin
      Inc(p);
      fName := FontName;
      Inc(fName, p);
      if StrComp('BoldItalic', fName) = 0 then
         Style := fsBold or fsItalic
      else if StrComp('Italic', fName) = 0 then
         Style := Style or fsItalic
      else if StrComp('Bold', fName) = 0 then
         Style := Style or fsBold;
      Dec(p);
      if p > 64 then p := 64;
      for i := 0 to p -1 do
         FamilyName[i] := FontName[i];
      FamilyName[p] := #0;
      Result := true;
   end else begin
      for i := 0 to Len -1 do
         FamilyName[i] := FontName[i];
      FamilyName[Len] := #0;
      if FindChar(FamilyName, ' ') > 0 then
         Result := true
      else
         Result := false;
   end;
end;

procedure CPDFEditText.CalcAlpha(var M: TCTM);
var x1, y1, x2, y2: Double;
begin
   x1 := 0.0;
   y1 := 0.0;
   x2 := 1.0;
   y2 := 0.0;
   Transform(M, x1, y1);
   Transform(M, x2, y2);
   // Get rid of rounding errors...
   m_Alpha := Trunc(ArcTan2(y2-y1, x2-x1) / 0.017453292519943295769236907684886 * 256.0);
end;

procedure CPDFEditText.CalcDistance(var M1, M2, M3: TCTM; var DistX, DistY: Double; x, y: Double);
begin
   if m_Alpha = 0 then begin
      DistX := x - m_LastX;
      DistY := abs(y - m_LastY);
   end else begin
      M3 := MulMatrix(M1, M2);
      {
         Notice: The distance is always a positive value. If the text is not ordered from left
         to right the result can be incorrect!

         To do:
         Check whether the text occurs behind the previous one. If not, insert the text at the
         beginning of the line. This is recommended if bidirectional text should be processed.
      }
      DistX := CalcDistance(m_LastX, m_LastY, x, y);
      DistY := abs(M3.y);
   end;
end;

function CPDFEditText.CalcDistance(x1, y1, x2, y2: Double): Double;
var dx, dy: Double;
begin
   dx := x2-x1;
   dy := y2-y1;
   Result := sqrt(dx*dx + dy*dy);
end;

procedure CPDFEditText.CalcStrPos(var M: TCTM; var x, y: Double);
begin
   x := 0.0;
   y := 0.0;
   M := MulMatrix(m_Stack.ctm, m_Stack.tm);
   Transform(M, x, y);
   CalcAlpha(M);
end;

procedure CPDFEditText.FindAll;
var m: TCTM; alpha: Integer; x, y, spaceWidth, distX, distY: Double;
begin
   m_LastX        := 0.0;
   m_LastY        := 0.0;
   m_RecordNumber := 0;
   m_NewLine      := true;
   m_SearchPos    := m_SearchText;
   // Get the first text record if any
   m_HaveMore := m_PDFInst.GetPageText(m_Stack);
   // No text found?
   if m_Stack.TextLen = 0 then Exit;

   // Calculate the text position and orientation angle
   CalcStrPos(m_Matrix1, x, y);

   FindPattern();

   spaceWidth :=  m_Stack.SpaceWidth * GetScaleFactor(m_Matrix1) * 0.5;

   // Calculate the end offset and invert the matrix.
   m_LastX := m_Stack.TextWidth;
   Transform(m_Matrix1, m_LastX, m_LastY);
   Invert(m_Matrix1);

   Inc(m_RecordNumber);

   if not m_HaveMore then Exit;

   alpha := m_Alpha;
   while m_PDFInst.GetPageText(m_Stack) do begin
      CalcStrPos(m_Matrix2, x, y);
      if alpha = m_Alpha then begin
         CalcDistance(m_Matrix1, m_Matrix2, m, distX, distY, x, y);
         if distY > MAX_LINE_ERROR then begin
            // This algorithm does not support line breaks!
            m_SearchPos := m_SearchText;
            m_Matrix1   := m_Matrix2;
            m_NewLine   := true;
            Invert(m_Matrix1);
         end else if distX > spaceWidth then begin
            if distX > 6.0 * spaceWidth then begin
               // The distance is too large. We assume that the text should not be considered
               // as part of the current text line.
               m_SearchPos := m_SearchText;
               m_Matrix1   := m_Matrix2;
               m_NewLine   := true;
               Invert(m_Matrix1);
            end else if m_SearchPos^ = #32 then begin
               if m_SearchPos = m_SearchText then begin
                  m_First      := m_RecordNumber;
                  m_KernRecord := -1;
                  m_StrPos     := -1;
               end;
               Inc(m_SearchPos);
               if m_SearchPos^ = #0 then begin
                  AddRecord(m_KernRecord, m_StrPos);
               end;
            end else begin
               m_SearchPos := m_SearchText;
               if m_SearchPos^ = #32 then begin
                  m_First := m_RecordNumber;
                  Inc(m_SearchPos);
                  if m_SearchPos^ = #0 then begin
                     AddRecord(-1, -1);
                  end;
               end;
            end;
         end else if distX < -spaceWidth then begin // Wrong direction?
            m_SearchPos := m_SearchText;
            m_Matrix1   := m_Matrix2;
            m_NewLine   := true;
            Invert(m_Matrix1);
         end;
      end else begin
         m_SearchPos := m_SearchText;
         m_Matrix1   := m_Matrix2;
         m_NewLine   := true;
         Invert(m_Matrix1);
      end;
      FindPattern();
      m_LastY    := 0.0;
      m_LastX    := m_Stack.TextWidth;
      alpha      := m_Alpha;
      spaceWidth := m_Stack.SpaceWidth * GetScaleFactor(m_Matrix2) * 0.5;
      Transform(m_Matrix2, m_LastX, m_LastY);
      Inc(m_RecordNumber);
   end;
end;

function CPDFEditText.FindChar(Text: PChar; AChar: Char): Integer;
begin
   // Do not use the function Pos() on a PChar!
   Result := 0;
   while Text^ <> #0 do begin
      if Text^ = AChar then Exit;
      Inc(Result);
      Inc(Text);
   end;
   Result := -1;
end;

function CPDFEditText.FindEndPattern(const Text: PWideChar; Len: Integer): Integer;
begin
   Result := 0;
   while Len > 0 do begin
      if m_SearchPos^ = #0 then Exit;
      if Text[Result] <> m_SearchPos^ then Exit;
      Inc(m_SearchPos);
      Inc(Result);
      Dec(Len);
   end;
end;

procedure CPDFEditText.FindPattern;
var i, j: Integer; c1, c2: WideChar; spw: Single; rec: TTextRecordWPtr;
begin
   spw := -m_Stack.SpaceWidth / 2.0;
   if m_SearchPos = m_SearchText then begin
      m_First      := m_RecordNumber;
      m_KernRecord := -1;
      m_StrPos     := -1;
   end;
   // Handle spaces
   rec := m_Stack.Kerning;
   for i := 0 to m_Stack.KerningCount -1 do begin
      if rec.Advance < spw then begin
         if m_SearchPos^ = #32 then begin
            if (m_SearchPos = m_SearchText) and (m_KernRecord < 0) then begin
               m_KernRecord := i;
               m_StrPos     := -1;
            end;
            Inc(m_SearchPos);
            if m_SearchPos^ = #0 then AddRecord(m_KernRecord, m_StrPos);
         end else begin
            m_SearchPos := m_SearchText;
            m_StrPos    := -1;
            if m_SearchPos^ = #32 then begin
               Inc(m_SearchPos);
               if m_KernRecord < 0 then m_KernRecord := i;
               if m_SearchPos^ = #0 then AddRecord(m_KernRecord, m_StrPos);
            end;
         end;
      end;
      // Compare the text.
      for j := 0 to rec.Length -1 do begin
         c1 := rec.Text[j];
         c2 := m_SearchPos^;
         if c1 <> c2 then begin
            case c1 of
               #160: begin // non-break space
                  if c2 = #32 then begin
                     if (m_SearchPos = m_SearchText) and (m_KernRecord < 0) then begin
                        m_KernRecord := i;
                        m_StrPos     := j;
                     end;
                     Inc(m_SearchPos);
                     continue;
                  end;
               end;
               #173: begin // non-break hyphen
                  if c2 = #45 then begin
                     if (m_SearchPos = m_SearchText) and (m_KernRecord < 0) then begin
                        m_KernRecord := i;
                        m_StrPos     := j;
                     end;
                     Inc(m_SearchPos);
                     continue;
                  end;
               end;
            end;
            m_KernRecord := -1;
            m_SearchPos  := m_SearchText;
            m_StrPos     := -1;
            continue;
         end else begin
            if (m_SearchPos = m_SearchText) and (m_KernRecord < 0) then begin
               m_KernRecord := i;
               m_StrPos     := j;
            end;
            Inc(m_SearchPos);
         end;
         if m_SearchPos^ = #0 then AddRecord(m_KernRecord, m_StrPos);
      end;
      Inc(rec);
   end;
end;

function CPDFEditText.FindText(const Text: WideString): Integer;
var i, len: Integer;
begin
   len := Length(Text);
   if len = 0 then raise Exception.Create('The search text cannot be an empty string!');
   if len > m_SearchTextLen then begin
      ReallocMem(m_SearchText, (len + 1) * sizeof(WideChar));
      m_SearchTextLen := len;
   end;
   for i := 0 to len -1 do
      m_SearchText[i] := Text[i+1];
   m_SearchText[len] := #0;
   m_LastFont := nil;
   m_Records.Clear();
   if not m_PDFInst.InitStack(m_Stack) then raise Exception.Create('-1');
   m_CurrTmpl := -1;
   FindAll();
   ParseTemplates();
   Result := m_Records.Count;
end;

function CPDFEditText.GetScaleFactor(var M: TCTM): Double;
var x, y: Double; const sin45: Double = 0.70710678118654752440084436210485;
begin
   x := sin45 * M.a + sin45 * M.c;
   y := sin45 * M.b + sin45 * M.d;
   Result := sqrt(x*x + y*y);
end;

procedure CPDFEditText.Invert(var M: TCTM);
var a, d, x: Double;
begin
   d   := 1.0 / (M.a * M.d - M.b * M.c);
   a   :=  M.d * d;
   M.d :=  M.a * d;
   M.b := -M.b * d;
   M.c := -M.c * d;
   x   := -M.x *   a - M.y * M.c;
   M.y := -M.x * M.b - M.y * M.d;
   M.a :=  a;
   M.x :=  x;
end;

function CPDFEditText.MulMatrix(var M1, M2: TCTM): TCTM;
begin
   Result.a := M2.a * M1.a + M2.b * M1.c;
   Result.b := M2.a * M1.b + M2.b * M1.d;
   Result.c := M2.c * M1.a + M2.d * M1.c;
   Result.d := M2.c * M1.b + M2.d * M1.d;
   Result.x := M2.x * M1.a + M2.y * M1.c + M1.x;
   Result.y := M2.x * M1.b + M2.y * M1.d + M1.y;
end;

procedure CPDFEditText.ParseTemplates;
var e, i, j, tmplCount, tmplCount2: Integer; found: Boolean;
begin
   found     := false;
   tmplCount := m_PDFInst.GetTemplCount;
   for i := 0 to tmplCount -1 do begin
      if not m_PDFInst.EditTemplate(i) then raise Exception.Create('-1');
      m_CurrTmpl := m_PDFInst.GetTemplHandle;
      // We must make sure that we don't parse a template twice
      j := 0;
      e := m_Templates.Count -1;
      while j <= e do begin
         if m_CurrTmpl = m_Templates.GetItem(j) then begin
            found := true;
            break;
         end;
         if m_CurrTmpl = m_Templates.GetItem(e) then begin
            found := true;
            break;
         end;
         Inc(j);
         Dec(e);
      end;
      if found then begin
         m_PDFInst.EndTemplate;
         continue;
      end;
      // Add the template handle to the list of templates
      m_Templates.Add(m_CurrTmpl);

      if not m_PDFInst.InitStack(m_Stack) then raise Exception.Create('-1');

      FindAll;

      tmplCount2 := m_PDFInst.GetTemplCount;
      for j := 0 to tmplCount2 -1 do begin
         ParseTemplates;
      end;
      m_PDFInst.EndTemplate;
   end;
end;

procedure CPDFEditText.ReplacePattern(const NewText: WideString);
var i, j, textLen, lastTmpl: Integer; rec: TTextRecPtr; textPtr: PWideChar; src: TTextRecordWPtr;
begin
   i        := 0; 
   textLen  := Length(NewText);
   textPtr  := PWideChar(NewText); 
   lastTmpl := -1;
   if not m_PDFInst.InitStack(m_Stack) then raise Exception.Create('-1');
	m_Matrix3.a := 1.0;
	m_Matrix3.b := 0.0;
	m_Matrix3.c := 0.0;
	m_Matrix3.d := 1.0;
	m_Matrix3.x := 0.0;
	m_Matrix3.y := 0.0;
   m_Matrix4.x := 0.0;
   m_RecordNumber := 0;
   while i < m_Records.Count do begin
      rec := m_Records.GetItem(i);
      if rec.TmplHandle <> lastTmpl then begin
         if not m_PDFInst.FlushPageContent(m_Stack) then raise Exception.Create('-1');
         if lastTmpl > -1 then m_PDFInst.EndTemplate;
         if not m_PDFInst.EditTemplate2(rec.TmplHandle) then raise Exception.Create('-1');
         if not m_PDFInst.InitStack(m_Stack) then raise Exception.Create('-1');
         m_RecordNumber := 0;
      end;
      while m_RecordNumber < rec.First do begin
         m_PDFInst.GetPageText(m_Stack);
         Inc(m_RecordNumber);
      end;
      if rec.NewLine then begin
         m_Matrix4.x := m_Stack.TextWidth;
         m_Matrix1 := MulMatrix(m_Stack.tm, m_Matrix4);
         m_Matrix3 := MulMatrix(m_Stack.ctm, m_Matrix1);
      end;
      m_HaveMore := m_PDFInst.GetPageText(m_Stack);
      if rec.KernRecord > 0 then begin
         // Delete the string but preserve the kerning records before the string occurred.
         m_Stack.DeleteKerningAt := rec.KernRecord;
         m_PDFInst.ReplacePageText('', m_Stack);
         m_Matrix4.x := 0.0;
         src := m_Stack.Kerning;
         for j := 0 to rec.KernRecord -1 do begin
            m_Matrix4.x := m_Matrix4.x - src.Advance + src.Width;
            Inc(src);
         end;
         // If StrPos == -1 the first character is a space character which is emulated with kerning space.
         if rec.StrPos > -1 then m_Matrix4.x := m_Matrix4.x - src.Advance;
         // Compute the string position
         m_Matrix1 := MulMatrix(m_Stack.tm, m_Matrix4);
         m_Matrix2 := MulMatrix(m_Stack.ctm, m_Matrix1);
         i := WriteLine(i, rec, textPtr, textLen);
      end else begin
         m_PDFInst.ReplacePageText('', m_Stack);
         // Compute the string position
         m_Matrix2 := MulMatrix(m_Stack.ctm, m_Stack.tm);
         i := WriteLine(i, rec, textPtr, textLen);
      end;
      lastTmpl := rec.TmplHandle;
   end;
   if not m_PDFInst.FlushPageContent(m_Stack) then raise Exception.Create('-1');
   if lastTmpl > -1 then m_PDFInst.EndTemplate();
end;

procedure CPDFEditText.SetFont;
var font: TPDFFontObj; familyName: Array[0..65] of Char; style: TFStyle;
begin
   if m_LastFont <> m_Stack.IFont then begin
      // Not that this function does not use the DynaPDF exception handling!
      // The only reason why this function can fail is out of memory.
      if not m_PDFInst.GetFont(m_Stack.IFont, font) then raise Exception.Create('Out of memory!');
      familyName[0] := #0;
      style         := fsNone;
      if (font.ItalicAngle <> 0.0) or (font.Flags and $40 <> 0) then
         style := style or fsItalic;
      // The force bold flag must be ignored if the font type is not Type1.
      if (font.FontType = ftType1) and (font.Flags and $40000 <> 0) then
         style := style or fsBold;
      BildFamilyNameAndStyle(font.FontName, StrLen(font.FontName), familyName, style);
      if m_PDFInst.SetFontExA(familyName, style, 10.0, true, cpUnicode) < 0 then begin
         if font.Flags and 1 <> 0 then begin // Fixed pitch
            if m_PDFInst.SetFontA('Courier New', style, 10.0, true, cpUnicode) < 0 then
               raise Exception.Create('-1');
         end else if m_PDFInst.SetFontA('Arial', style, 10.0, true, cpUnicode) < 0 then
            raise Exception.Create('-1');
      end;
      m_LastFont := m_Stack.IFont;
   end;
end;

procedure CPDFEditText.Transform(var M: TCTM; var x, y: Double);
var tx: Double;
begin
   tx := x;
   x  := tx * M.a + y * M.c + M.x;
   y  := tx * M.b + y * M.d + M.y;
end;

function CPDFEditText.WriteLine(Index: Integer; TextRecord: TTextRecPtr; NewText: PWideChar; Len: Integer): Integer;
var i, j, p, count: Integer; rec: TTextRecPtr; src: TTextRecordWPtr; spaceWidth: Single;
begin
   // Process the remaining text in the kerning array
   Inc(Index);
   count := 0;
   spaceWidth := -m_Stack.SpaceWidth * 0.5;
   while Index < m_Records.Count do begin
      rec := m_Records.GetItem(Index);
      if (rec.First <> TextRecord.First) or (rec.TmplHandle <> TextRecord.TmplHandle) then break;
      Inc(Index);
      Inc(count);
   end;
   {
      We must always handle the text before and after the search text was found.
      Handling space characters correctly is the most important part if text
      should be replaced or deleted.
   }
   if count > 0 then begin
      // The text was found multiple times in the text record
      src := m_Stack.Kerning;
      Inc(src, TextRecord.KernRecord);
      WriteRecord(m_Matrix2, TextRecord, src, NewText, Len);
      Dec(Index, count);
      while count > 0 do begin
         Dec(count);
         rec := m_Records.GetItem(Index);
         for j := TextRecord.KernRecord + 1 to rec.KernRecord -1 do begin
            Inc(src);
            AddKernSpace(src.Advance, spaceWidth, m_Matrix2);
            WriteText(m_Matrix2, src.Text, src.Length);
         end;
         Inc(src);
         if (rec.StrPos > -1) and (m_SearchPos^ <> #32) then begin
            m_Matrix4.x := -src.Advance;
            m_Matrix2   := MulMatrix(m_Matrix2, m_Matrix4);
         end;
         WriteRecord(m_Matrix2, rec, src, NewText, Len);
         TextRecord := rec;
         Inc(Index);
      end;
      // Handle the remaining records in the kerning array
      for i := TextRecord.KernRecord + 1 to  m_Stack.KerningCount -1 do begin
         Inc(src);
         AddKernSpace(src.Advance, spaceWidth, m_Matrix2);
         WriteText(m_Matrix2, src.Text, src.Length);
      end;
   end else if TextRecord.KernRecord > -1 then begin
      src := m_Stack.Kerning;
      Inc(src, TextRecord.KernRecord);
      WriteRecord(m_Matrix2, TextRecord, src, NewText, Len);
      for i := TextRecord.KernRecord + 1 to m_Stack.KerningCount -1 do begin
         Inc(src);
         AddKernSpaceEx(src.Advance, spaceWidth, m_Matrix2);
         if m_SearchPos^ <> #0 then begin
            p := FindEndPattern(src.Text, src.Length);
            WriteText(m_Matrix2, src.Text + p, src.Length - p);
         end else
            WriteText(m_Matrix2, src.Text, src.Length);
      end;
   end else begin
      // The text begins at the end of the previous text record.
      src := m_Stack.Kerning;
      WriteRecord(m_Matrix3, TextRecord, src, NewText, Len);
      for i := 1 to m_Stack.KerningCount -1 do begin
         Inc(src);
         AddKernSpace(src.Advance, spaceWidth, m_Matrix3);
         WriteText(m_Matrix3, src.Text, src.Length);
      end;
      m_Matrix2 := m_Matrix3;
   end;
   Result := WriteRemaining(Index, TextRecord, NewText, Len);
end;

procedure CPDFEditText.WriteRecord(var M: TCTM; TextRecord: TTextRecPtr; Source: TTextRecordWPtr; NewText: PWideChar; Len: Integer);
var p, sl: Integer;
begin
   m_SearchPos := m_SearchText;
   if TextRecord.StrPos < 0 then begin
      Inc(m_SearchPos);
      p  := FindEndPattern(Source.Text, Source.Length);
      sl := Source.Length - p;
      // Write the new text if any
      if Len > 0 then WriteText(M, NewText, Len);
      if sl > 0 then WriteText(M, Source.Text + p, sl);
   end else begin
      p := TextRecord.StrPos;
      if p > 0 then begin
         WriteText(M, Source.Text, p);
         // Write the new text if any
         if Len > 0 then WriteText(M, NewText, Len);
         Inc(p, FindEndPattern(Source.Text + p, Source.Length - p));
         if p < Source.Length then WriteText(M, Source.Text + p, Source.Length - p);
      end else begin
         p  := FindEndPattern(Source.Text, Source.Length);
         sl := Source.Length - p;
         if sl > 0 then WriteText(M, Source.Text + p, sl);
         // Write the new text if any
         if Len > 0 then WriteText(M, NewText, Len);
      end;
   end;
end;

function CPDFEditText.WriteRemaining(Index: Integer; TextRecord: TTextRecPtr; NewText: PWideChar; Len: Integer): Integer;
var i, j, p, sl, lastAlpha, lastRecord: Integer; m1, m2: TCTM; rec: TTextRecPtr;
spaceWidth1: Single; x, y, spaceWidth2, distX, distY: Double; src: TTextRecordWPtr;
begin
   {
		We handle here the rest of the text line if it is stored in separate text records.
		This is rather complex especially if the search text was found multiple times in the
		line and if the line consists in turn of multiple text records. We handle
		this case recursivley because this is much easier in comparison to a non-recursive
		version.
   }
   Result := Index;
   rec := nil;
   if not m_HaveMore then begin
      Inc(m_RecordNumber);
      Exit;
   end;
   if Result >= m_Records.Count then
      lastRecord := $7FFFFFFF
   else begin
      rec        := m_Records.GetItem(Result);
      lastRecord := rec.First;
   end;
   m_LastX   := 0.0;
   m_LastY   := 0.0;
   m_Matrix3 := m_Matrix2;
   // Get the end offset of the last text record. The current position of the
   // original text is different because the new text was already written to
   // the file.
   m_Matrix4.x := m_Stack.TextWidth;
   m_Matrix2   := MulMatrix(m_Stack.tm, m_Matrix4);
   m_Matrix2   := MulMatrix(m_Stack.ctm, m_Matrix2);
   Transform(m_Matrix2, m_LastX, m_LastY);
   CalcAlpha(m_Matrix2);
   spaceWidth2 := m_Stack.SpaceWidth * GetScaleFactor(m_Matrix2) * 0.5;
   Inc(m_RecordNumber);
   // Was the search string found in multiple records?
   while m_RecordNumber < TextRecord.Last do begin
      m_HaveMore := m_PDFInst.GetPageText(m_Stack);
      if not m_HaveMore then Exit;
      Inc(m_RecordNumber);
      spaceWidth1 := -m_Stack.SpaceWidth / 2.0;
      m_Matrix2   := MulMatrix(m_Stack.ctm, m_Stack.tm);
      x := 0.0;
      y := 0.0;
      Transform(m_Matrix2, x, y);
      AddSpace(x, y, spaceWidth2);
      m_PDFInst.ReplacePageText('', m_Stack);
      src := m_Stack.Kerning;
      for i := 0 to m_Stack.KerningCount -1 do begin
         AddKernSpaceEx(src.Advance, spaceWidth1, m_Matrix3);
         if m_SearchPos^ <> #0 then begin
            p  := FindEndPattern(src.Text, src.Length);
            sl := src.Length - p;
            if sl > 0 then begin
               // We are now at the end of the already replaced text. Only the remaining text behind the
               // search text must be written to the file.
               WriteText(m_Matrix3, src.Text + p, sl);
            end;
         end else
            WriteText(m_Matrix3, src.Text, src.Length);
         Inc(src);
      end;
      spaceWidth2 := m_Stack.SpaceWidth * GetScaleFactor(m_Matrix2) * 0.5;
      m_LastY     := 0.0;
      m_LastX     := m_Stack.TextWidth;
      Transform(m_Matrix2, m_LastX, m_LastY);
   end;
   {
      Now we come to the most complicated part: we must find all records which lie on the same
      text line and move the text behind the replaced one to avoid overlaping text or holes
      between text records. We must consider the spacing between records to preserve the original
      layout. Note that we deal with two text positions here: m_Matrix2 contains the coordinates
      of the original text while m_Matrix3 represents the position of the new text.
   }
   m1 := m_Matrix2;
   Invert(m1);
   CalcAlpha(m_Matrix2);
   lastAlpha := m_Alpha;
   while m_HaveMore and (m_RecordNumber < lastRecord) do begin
      m_HaveMore := m_PDFInst.GetPageText(m_Stack);
      if not m_HaveMore then Exit;
      Inc(m_RecordNumber);
      spaceWidth1 := -m_Stack.SpaceWidth / 2.0;
      CalcStrPos(m_Matrix2, x, y);
      if m_Alpha <> lastAlpha then begin
         rec := nil;
         break;
      end;
      CalcDistance(m1, m_Matrix2, m2, distX, distY, x, y);
      if distY > MAX_LINE_ERROR then begin
         // We are on a new text line...
         rec := nil;
         break;
      end else if (distX < -spaceWidth2) or (distX > 6.0 * spaceWidth2) then begin
         // The distance is too large. We assume that the text should not be considered
         // as part of the current text line.
         rec := nil;
         break;
      end;
      m_PDFInst.ReplacePageText('', m_Stack);     
      AddSpace(distX, spaceWidth2, m_Matrix3);
      src := m_Stack.Kerning;
      for i := 0 to m_Stack.KerningCount -1 do begin
         AddKernSpaceEx(src.Advance, spaceWidth1, m_Matrix3);
         if m_SearchPos^ <> #0 then begin
            p  := FindEndPattern(src.Text, src.Length);
            sl := src.Length - p;
            if sl > 0 then begin
               // We are now at the end of the already replaced text. Only the remaining text behind the
               // search text must be written to the file.
               WriteText(m_Matrix3, src.Text + p, sl);
            end;
         end else
            WriteText(m_Matrix3, src.Text, src.Length);
         Inc(src);
      end;
      lastAlpha   := m_Alpha;
      spaceWidth2 := m_Stack.SpaceWidth * GetScaleFactor(m_Matrix2) * 0.5;
      m_LastY     := 0.0;
      m_LastX     := m_Stack.TextWidth;
      Transform(m_Matrix2, m_LastX, m_LastY);
   end;
   if (rec <> nil) and not rec.NewLine then begin
      Inc(Result);
      m_HaveMore := m_PDFInst.GetPageText(m_Stack);
      // We are still in the same text line. We must check whether a space character must be deleted.
      // The rest of the line is now processed recursively.
      CalcStrPos(m_Matrix2, x, y);
      if m_Alpha = 0 then
         distX := x - m_LastX
      else
         distX := CalcDistance(m_LastX, m_LastY, x, y);
      AddSpace(distX, spaceWidth2, m_Matrix3);
      if rec.KernRecord > 0 then begin
         // Delete the string but preserve the kerning records before the string occurred.
         m_Stack.DeleteKerningAt := rec.KernRecord;
         m_PDFInst.ReplacePageText('', m_Stack);
         m_Matrix4.x := 0.0;
         src := m_Stack.Kerning;
         for j := 0 to rec.KernRecord -1 do begin
            m_Matrix4.x := m_Matrix4.x - src.Advance + src.Width;
            Inc(src);
         end;
         // If StrPos == -1 the first character is a space character which is emulated with kerning space.
         if rec.StrPos > -1 then m_Matrix4.x := m_Matrix4.x - src.Advance;
         // Compute the string position
         m_Matrix2 := MulMatrix(m_Matrix3, m_Matrix4);
         Result := WriteLine(Result, rec, NewText, Len);
      end else begin
         m_PDFInst.ReplacePageText('', m_Stack);
         m_Matrix2 := m_Matrix3;
         Result := WriteLine(Result, rec, NewText, Len);
      end;
   end;
end;

procedure CPDFEditText.WriteText(var M: TCTM; Text: PWideChar; Len: Integer);
begin
   // Initialize the graphics state
   SetFont;
   // The new text is written in red color so that you can better find it
   m_PDFInst.SetFillColorSpace(m_Stack.FillCS);
   m_PDFInst.SetFillColor(255);
   m_PDFInst.SetTextDrawMode(m_Stack.DrawMode);
{
   // This is the normal initalization
   m_PDFInst.SetFillColorSpace(m_Stack.FillCS);
   m_PDFInst.SetFillColor(m_Stack.FillColor);
   m_PDFInst.SetStrokeColorSpace(m_Stack.StrokeCS);
   m_PDFInst.SetStrokeColor(m_Stack.StrokeColor);
   m_PDFInst.SetTextDrawMode(m_Stack.DrawMode);
}
   m_PDFInst.ChangeFontSize(m_Stack.FontSize);
   m_PDFInst.WriteTextMatrixExW(M, Text, Len);
   // Compute the end offset of the text
   m_Matrix4.x := m_PDFInst.GetTextWidthExW(Text, Len);
   M := MulMatrix(M, m_Matrix4);
end;

end.
