using System;
using System.Runtime.InteropServices;
using DynaPDF;

namespace PDFUtils
{
   /// <summary>
   /*
      This class demonstrates how texts can be found and replaced in a PDF file. The
      algorithm handles rotated text as well as text lines which consist of multiple
      text records correctly.

      This algorithm assumes that the text is ordered from left to right and from top
      to bottom. If this is not the case the result is maybe incorrect or the search
      text cannot be found.

		The remaing text in a line is realigned to avoid overlaping text or holes
		between text redords. This realignment is not a real text re-flow because it is
		restricted to the current text line. No line break occurs if the new text is too
		large to fit into the page or text block. It is possible to develop a real text
		re-flow but such an algorithm is much more complex than this one.

		A page must always be parsed twice, the first time to find the text records which
		make up the search string, and a second time to delete or replace the text.

		----------------------------------------------------------------------------------

		A note about underlined text:
		The underline of a text is not a font attribute. It is a separate vector graphic
		which occurs outside of the text object, e.g. a rectanlge or a line drawn with a
		moveto/lineto pair. DynaPDF contains no algorithms to identify such underlines so
		that they can be deleted with the text. If you delete an underlined text the
		underline will be left unchanged and only the text is deleted!
   */

   /*
      Calculation of the space width:

      Space characters are often not stored in PDF files, instead, kerning space
      is applied which produces the same visual appearance. Due to the missing space
      characters we need an algorithm that is able to detect word boundaries.

      The algorithm here compares just the distance between two text records and the
      the displacement in kerning records. If it is larger than the half space width
      we assume that a space character was emulated at this position. Note that the
      displacement in a kerning record is measured in text space while the distance
      between two text records is measured in user space!

      The full space width is often too large to determine whether normal kerning
      space was applied or whether a space character must be added at a given position.
      The half space width procudes better results in most cases especially if the
      document uses condensed fonts. The fonts of documents which emulate space
      characters contain often no space character. In this case a standard space width
      is set because the real value is not known.
   */
   /// </summary>

   public struct TTextRec
   {
      public int  TmplHandle;
      public int  First;
      public int  Last;
      public int  KernRecord;
      public int  StrPos;
      public bool NewLine;
   }

	internal class CPDFEditText
	{
      // Allowed error to determine whether a text record must be assigned to the current line.
      private const double MAX_LINE_ERROR = 2.0;

      // List to store found text records
      public class CTextRecords
      {
         public void Add(int TmplHandle, int First, int Last, int KernRecord, int StrPos, bool NewLine)
         {
            if (m_Count + 1 > m_Capacity)
            {
               TTextRec[] tmp = new TTextRec[m_Capacity + 4196];
               if (m_Items != null) m_Items.CopyTo(tmp, 0);
               m_Items = tmp;
               tmp = null;
               m_Capacity += 4196;
            }
            m_Items[m_Count].TmplHandle = TmplHandle;
            m_Items[m_Count].First      = First;
            m_Items[m_Count].Last       = Last;
            m_Items[m_Count].KernRecord = KernRecord;
            m_Items[m_Count].StrPos     = StrPos;
            m_Items[m_Count].NewLine    = NewLine;
            ++m_Count;
         }
         public void Clear()                {m_Count = 0;}
         public int  Count()                {return m_Count;}
         public TTextRec GetItem(int Index) {return m_Items[Index];}

         private int        m_Capacity;
         private int        m_Count;
         private TTextRec[] m_Items;
      }
      // List to store template handles
      public class CIntList
      {
         public void Add(int Value)
         {
            if (m_Count + 1 > m_Capacity)
            {
               int[] tmp = new int[m_Capacity + 1024];
               if (m_Items != null) m_Items.CopyTo(tmp, 0);
               m_Items = tmp;
               tmp = null;
               m_Capacity += 1024;
            }
            m_Items[m_Count++] = Value;
         }
         public void Clear()                {m_Count = 0;}
         public int  Count()                {return m_Count;}
         public int  GetItem(int Index) {return m_Items[Index];}

         private int   m_Capacity;
         private int   m_Count;
         private int[] m_Items;
      }

      internal CPDFEditText(CPDF PDFInst)
      {
         m_Matrix4.a = 1.0;
         m_Matrix4.d = 1.0;
         m_PDFInst   = PDFInst;
         m_Records   = new CTextRecords();
         m_Templates = new CIntList();
      }
      private void AddKernSpace(float Advance, float SpaceWidth, ref TCTM Matrix)
      {
         if (Advance < SpaceWidth)
         {
            if (m_SearchPos < m_SearchText.Length && m_SearchText[m_SearchPos] == 32)
            {
               ++m_SearchPos;
               if (m_SearchPos == m_SearchText.Length) m_SearchPos = 0;
            }else
            {
               m_Matrix4.x = -Advance;
               Matrix      = MulMatrix(ref Matrix, ref m_Matrix4);
            }
         }else
         {
            m_Matrix4.x = -Advance;
            Matrix      = MulMatrix(ref Matrix, ref m_Matrix4);
         }
      }
      private void AddKernSpaceEx(float Advance, float SpaceWidth, ref TCTM Matrix)
      {
         if (Advance < SpaceWidth)
         {
            if (m_SearchPos < m_SearchText.Length && m_SearchText[m_SearchPos] == 32)
               ++m_SearchPos;
            else
            {
               m_Matrix4.x = -Advance;
               Matrix      = MulMatrix(ref Matrix, ref m_Matrix4);
            }
         }else
         {
            m_Matrix4.x = -Advance;
            Matrix      = MulMatrix(ref Matrix, ref m_Matrix4);
         }
      }
      private void AddRecord(int KernRecord, int StrPos)
      {
         m_Records.Add(m_CurrTmpl, m_First, m_RecordNumber, KernRecord, StrPos, m_NewLine);
         m_KernRecord   = -1;
         m_NewLine      = false;
         m_SearchPos    = 0;
         m_StrPos       = -1;
      }
      private void AddSpace(double x, double y, double SpaceWidth)
      {
         double distX;
         if (m_Alpha == 0)
            distX = x - m_LastX;
         else
            distX = CalcDistance(m_LastX, m_LastY, x, y);
         if (distX > SpaceWidth)
         {
            if (m_SearchPos < m_SearchText.Length && m_SearchText[m_SearchPos] == 32)
            {
               ++m_SearchPos;
            }
         }
      }
      private void AddSpace(double DistX, double SpaceWidth, ref TCTM Matrix)
      {
         if (DistX > SpaceWidth)
         {
            if (m_SearchPos < m_SearchText.Length && m_SearchText[m_SearchPos] == 32)
               ++m_SearchPos;
            else
            {
               m_Matrix4.x = DistX;
               Matrix      = MulMatrix(ref Matrix, ref m_Matrix4);
            }
         }else
         {
            m_Matrix4.x = DistX;
            Matrix      = MulMatrix(ref Matrix, ref m_Matrix4);
         }
      }
      private bool BildFamilyNameAndStyle(String FontName, ref String FamilyName, ref DynaPDF.TFStyle Style)
      {
         int first = 0, p;
         if (FontName.Length > 7 && FontName[6] == '+')
         {
            first = 7;
         }
         if ((p = FontName.IndexOf(',', first)) > -1)
         {
            ++p;
            String fstyle = FontName.Substring(p);
            if (fstyle.IndexOf("BoldItalic") > -1)
               Style |= TFStyle.fsBold | TFStyle.fsItalic;
            else if (fstyle.IndexOf("Italic") > -1)
               Style |= TFStyle.fsItalic;
            else if (fstyle.IndexOf("Bold") > -1)
               Style |= TFStyle.fsBold;
            --p;
            FamilyName = FontName.Substring(first, p);
            return true;
         }else
         {
            FamilyName = FontName.Substring(first);
            if (FamilyName.IndexOf(' ', 0) > -1) return true;
            return false;
         }
      }
      private void CalcAlpha(ref TCTM M)
      {
         double x1 = 0.0;
         double y1 = 0.0;
         double x2 = 1.0;
         double y2 = 0.0;
         Transform(ref M, ref x1, ref y1);
         Transform(ref M, ref x2, ref y2);
         // Get rid of rounding errors...
         m_Alpha = (int)((Math.Atan2(y2-y1, x2-x1) / 0.017453292519943295769236907684886) * 256.0);
      }
      private void CalcDistance(ref TCTM M1, ref TCTM M2, ref TCTM M3, ref double DistX, ref double DistY, double x, double y)
      {
         if (m_Alpha == 0)
         {
            DistX = x - m_LastX;
            DistY = Math.Abs(y - m_LastY);
         }else
         {
            M3 = MulMatrix(ref M1, ref M2);
            /*
               Notice: The distance is always a positive value. If the text is not ordered from left
               to right the result can be incorrect!

               To do:
               Check whether the text occurs behind the previous one. If not, insert the text at the
               beginning of the line. This is recommended if bidirectional text should be processed.
            */
            DistX = CalcDistance(m_LastX, m_LastY, x, y);
            DistY = Math.Abs(M3.y);
         }
      }
      private double CalcDistance(double x1, double y1, double x2, double y2)
      {
         double dx = x2-x1;
         double dy = y2-y1;
         return Math.Sqrt(dx*dx + dy*dy);
      }
      private void CalcStrPos(ref TCTM M, ref double x, ref double y)
      {
         x = 0.0;
         y = 0.0;
         M = MulMatrix(ref m_Stack.ctm, ref m_Stack.tm);
         Transform(ref M, ref x, ref y);
         CalcAlpha(ref M);
      }
      private void FindAll()
      {
         TCTM m = new TCTM();
         int alpha;
         double x = 0.0, y = 0.0, spaceWidth, distX = 0.0, distY = 0.0;
         m_LastX        = 0.0;
         m_LastY        = 0.0;
         m_RecordNumber = 0;
         m_NewLine      = true;
         m_SearchPos    = 0;
         // Get the first text record if any
         m_HaveMore = m_PDFInst.GetPageText(ref m_Stack);
         // No text found?
         if (m_Stack.TextLen == 0) return;

         // Calculate the text position and orientation angle
         CalcStrPos(ref m_Matrix1, ref x, ref y);

         FindPattern();

         spaceWidth =  m_Stack.SpaceWidth * GetScaleFactor(ref m_Matrix1) * 0.5;

         // Calculate the end offset and invert the matrix.
         m_LastX = m_Stack.TextWidth;
         Transform(ref m_Matrix1, ref m_LastX, ref m_LastY);
         Invert(ref m_Matrix1);

         ++m_RecordNumber;

         if (!m_HaveMore) return;

         alpha = m_Alpha;
         while (m_PDFInst.GetPageText(ref m_Stack))
         {
            CalcStrPos(ref m_Matrix2, ref x, ref y);
            if (alpha == m_Alpha)
            {
               CalcDistance(ref m_Matrix1, ref m_Matrix2, ref m, ref distX, ref distY, x, y);
               if (distY > MAX_LINE_ERROR)
               {
                  // This algorithm does not support line breaks!
                  m_SearchPos = 0;
                  m_Matrix1   = m_Matrix2;
                  Invert(ref m_Matrix1);
                  m_NewLine = true;
               }else if (distX > spaceWidth)
               {
                  if (distX > 6.0 * spaceWidth)
                  {
                     // The distance is too large. We assume that the text should not be considered
                     // as part of the current text line.
                     m_SearchPos = 0;
                     m_Matrix1   = m_Matrix2;
                     Invert(ref m_Matrix1);
                     m_NewLine = true;
                  }else if (m_SearchText[m_SearchPos] == ' ')
                  {
                     if (m_SearchPos == 0)
                     {
                        m_First      = m_RecordNumber;
                        m_KernRecord = -1;
                        m_StrPos     = -1;
                     }
                     ++m_SearchPos;
                     if (m_SearchPos == m_SearchText.Length)
                     {
                        AddRecord(m_KernRecord, m_StrPos);
                     }
                  }else
                  {
                     m_SearchPos = 0;
                     if (m_SearchText[m_SearchPos] == 32)
                     {
                        m_First = m_RecordNumber;
                        ++m_SearchPos;
                        if (m_SearchPos == m_SearchText.Length)
                        {
                           AddRecord(-1, -1);
                        }
                     }
                  }
               }else if (distX < -spaceWidth)
               {
                  m_SearchPos = 0;
                  m_Matrix1   = m_Matrix2;
                  Invert(ref m_Matrix1);
                  m_NewLine = true;
               }
            }else
            {
               m_SearchPos = 0;
               m_Matrix1   = m_Matrix2;
               Invert(ref m_Matrix1);
               m_NewLine = true;
            }
            FindPattern();
            m_LastY    = 0.0;
            m_LastX    = m_Stack.TextWidth;
            alpha      = m_Alpha;
            spaceWidth = m_Stack.SpaceWidth * GetScaleFactor(ref m_Matrix2) * 0.5;
            Transform(ref m_Matrix2, ref m_LastX, ref m_LastY);
            ++m_RecordNumber;
         }
      }
      private int FindEndPattern(String Text, int Start)
      {
         int len = Text.Length - Start;
         while (len-- > 0)
         {
            if (m_SearchPos == m_SearchText.Length) break;
            if (Text[Start] != m_SearchText[m_SearchPos]) break;
            ++m_SearchPos;
            ++Start;
         }
         return Start;
      }
      private void FindPattern()
      {
         int i, j;
         char c1, c2;
         long ptr;
         String text;
         TTextRecordW rec = new TTextRecordW();
         float spw = -m_Stack.SpaceWidth / 2.0f;
         if (m_SearchPos == 0)
         {
            m_First      = m_RecordNumber;
            m_KernRecord = -1;
            m_StrPos     = -1;
         }
         // Handle spaces
         ptr = (long)m_Stack.Kerning;
         for (i = 0; i < m_Stack.KerningCount; i++)
         {
            // .Net languages are too stupid to handle arrays! We must copy each record manually.
            CPDF.CopyKernRecord(new IntPtr(ptr), ref rec, Marshal.SizeOf(rec));
            text = Marshal.PtrToStringUni(rec.Text, rec.Length);
            ptr += Marshal.SizeOf(rec);
            if (rec.Advance < spw)
            {
               if (m_SearchText[m_SearchPos] == ' ')
               {
                  if (m_SearchPos == 0 && m_KernRecord < 0)
                  {
                     m_KernRecord = i;
                     m_StrPos     = -1;
                  }
                  ++m_SearchPos;
                  if (m_SearchPos == m_SearchText.Length) AddRecord(m_KernRecord, m_StrPos);
               }else
               {
                  m_SearchPos = 0;
                  m_StrPos    = -1;
                  if (m_SearchText[m_SearchPos] == ' ')
                  {
                     ++m_SearchPos;
                     if (m_KernRecord < 0) m_KernRecord = i;
                     if (m_SearchPos == m_SearchText.Length) AddRecord(m_KernRecord, m_StrPos);
                  }
               }
            }
            // This is a case-sensitive search.
            for (j = 0; j < text.Length; j++)
            {
               c1 = text[j];
               c2 = m_SearchText[m_SearchPos];
               if (c1 != c2)
               {
                  switch(c1)
                  {
                     case (char)160: // non-break space
                     {
                        if (c2 == ' ')
                        {
                           if (m_SearchPos == 0 && m_KernRecord < 0)
                           {
                              m_KernRecord = i;
                              m_StrPos     = j;
                           }
                           ++m_SearchPos;
                           continue;
                        }
                        break;
                     }
                     case (char)173: // non-break hyphen
                     {
                        if (c2 == '-')
                        {
                           if (m_SearchPos == 0 && m_KernRecord < 0)
                           {
                              m_KernRecord = i;
                              m_StrPos     = j;
                           }
                           ++m_SearchPos;
                           continue;
                        }
                        break;
                     }
                     default:  break;
                  }
                  m_KernRecord = -1;
                  m_SearchPos  = 0;
                  m_StrPos     = -1;
                  continue;
               }else
               {
                  if (m_SearchPos == 0 && m_KernRecord < 0)
                  {
                     m_KernRecord = i;
                     m_StrPos     = j;
                  }
                  ++m_SearchPos;
               }
               if (m_SearchPos == m_SearchText.Length) AddRecord(m_KernRecord, m_StrPos);
            }
         }
      }
      public int FindPattern(String Text)
      {
         if (Text.Length == 0) throw new Exception("The search text cannot be an empty string!");
         m_SearchText = Text;
         m_Records.Clear();
         if (!m_PDFInst.InitStack(ref m_Stack)) throw new Exception("-1");
         m_CurrTmpl = -1;
         FindAll();
         ParseTemplates();
         return m_Records.Count();
      }
      private double GetScaleFactor(ref TCTM M)
      {
         double x = 0.70710678118654752440084436210485 * M.a + 0.70710678118654752440084436210485 * M.c;
         double y = 0.70710678118654752440084436210485 * M.b + 0.70710678118654752440084436210485 * M.d;
         return Math.Sqrt(x*x + y*y);
      }
      private void Invert(ref TCTM M)
      {
         double  d = 1.0 / (M.a * M.d - M.b * M.c);
         double  a =  M.d * d;
               M.d =  M.a * d;
               M.b = -M.b * d;
               M.c = -M.c * d;
                 d = -M.x *   a - M.y * M.c;
               M.y = -M.x * M.b - M.y * M.d;
               M.a =  a;
               M.x =  d;
      }
      private TCTM MulMatrix(ref TCTM M1, ref TCTM M2)
      {
         TCTM retval;
         retval.a = M2.a * M1.a + M2.b * M1.c;
         retval.b = M2.a * M1.b + M2.b * M1.d;
         retval.c = M2.c * M1.a + M2.d * M1.c;
         retval.d = M2.c * M1.b + M2.d * M1.d;
         retval.x = M2.x * M1.a + M2.y * M1.c + M1.x;
         retval.y = M2.x * M1.b + M2.y * M1.d + M1.y;
         return retval;
      }
      private void ParseTemplates()
      {
         bool found = false;
         int e, i, j, tmplCount, tmplCount2;
         tmplCount = m_PDFInst.GetTemplCount();
         for (i = 0; i < tmplCount; i++)
         {
            if (!m_PDFInst.EditTemplate(i)) throw new Exception("-1");
            m_CurrTmpl = m_PDFInst.GetTemplHandle();
            // We must make sure that we don't parse a template twice
            j = 0;
            e = m_Templates.Count() -1;
            while (j <= e)
            {
               if (m_Templates.GetItem(j) == m_CurrTmpl)
               {
                  found = true;
                  break;
               }
               if (m_Templates.GetItem(e) == m_CurrTmpl)
               {
                  found = true;
                  break;
               }
               ++j;
               --e;
            }
            if (found)
            {
               m_PDFInst.EndTemplate();
               continue;
            }
            // Add the template handle to the list of templates
            m_Templates.Add(m_CurrTmpl);

            if (!m_PDFInst.InitStack(ref m_Stack)) throw new Exception("-1");

            FindAll();

            tmplCount2 = m_PDFInst.GetTemplCount();
            for (j = 0; j < tmplCount2; j++)
            {
               ParseTemplates();
            }
            m_PDFInst.EndTemplate();
         }
      }
      public void ReplacePattern(String NewText)
      {
         int i = 0, j, lastTmpl = -1;
         long ptr;
         TTextRec rec;
         TTextRecordW src = new TTextRecordW();
         if (!m_PDFInst.InitStack(ref m_Stack)) throw new Exception("-1");
         m_Matrix3.a = 1.0;
         m_Matrix3.b = 0.0;
         m_Matrix3.c = 0.0;
         m_Matrix3.d = 1.0;
         m_Matrix3.x = 0.0;
         m_Matrix3.y = 0.0;
         m_RecordNumber = 0;
         m_LastFont = IntPtr.Zero;
         while (i < m_Records.Count())
         {
            rec = m_Records.GetItem(i);
            if (rec.TmplHandle != lastTmpl)
            {
               if (!m_PDFInst.FlushPageContent(ref m_Stack)) throw new Exception("-1");
               if (lastTmpl > -1) m_PDFInst.EndTemplate();
               if (!m_PDFInst.EditTemplate2(rec.TmplHandle)) throw new Exception("-1");
               if (!m_PDFInst.InitStack(ref m_Stack)) throw new Exception("-1");
               m_LastFont     = IntPtr.Zero;
               m_RecordNumber = 0;
            }
            while (m_RecordNumber < rec.First)
            {
               m_PDFInst.GetPageText(ref m_Stack);
               ++m_RecordNumber;
            }
            if (rec.NewLine)
            {
               m_Matrix4.x = m_Stack.TextWidth;
               m_Matrix1   = MulMatrix(ref m_Stack.tm, ref m_Matrix4);
               m_Matrix3   = MulMatrix(ref m_Stack.ctm, ref m_Matrix1);
            }
            m_HaveMore = m_PDFInst.GetPageText(ref m_Stack);
            if (rec.KernRecord > 0)
            {
               // Delete the string but preserve the kerning records before the string occurred.
               m_Stack.DeleteKerningAt = rec.KernRecord;
               m_PDFInst.ReplacePageTextA(null, ref m_Stack);
               m_Matrix4.x = 0.0;
               ptr = (long)m_Stack.Kerning;
               for (j = 0; j < rec.KernRecord; j++)
               {
                  CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
                  m_Matrix4.x -= src.Advance;
                  m_Matrix4.x += src.Width;
                  ptr         += Marshal.SizeOf(src);
               }
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               // If StrPos == -1 the first character is a space character which is emulated with kerning space.
               if (rec.StrPos > -1) m_Matrix4.x -= src.Advance;
               // Compute the string position
               m_Matrix1 = MulMatrix(ref m_Stack.tm, ref m_Matrix4);
               m_Matrix2 = MulMatrix(ref m_Stack.ctm, ref m_Matrix1);
               i = WriteLine(i, ref rec, NewText);
            }else
            {
               m_PDFInst.ReplacePageTextA(null, ref m_Stack);
               // Compute the string position
               m_Matrix2 = MulMatrix(ref m_Stack.ctm, ref m_Stack.tm);
               i = WriteLine(i, ref rec, NewText);
            }
            lastTmpl = rec.TmplHandle;
         }
         if (!m_PDFInst.FlushPageContent(ref m_Stack)) throw new Exception("-1");
         if (lastTmpl > -1) m_PDFInst.EndTemplate();
      }
      private void SetFont()
      {
         if (m_LastFont != m_Stack.IFont)
         {
            TPDFFontObj font = new TPDFFontObj();
            // Note that this function does not use the DynaPDF exception handling!
            // The only reason why this function can fail is out of memory.
            if (!m_PDFInst.GetFont(m_Stack.IFont, ref font)) throw new Exception("Out of memory!");
            String familyName = null;
            TFStyle style = TFStyle.fsNone;
            if (font.ItalicAngle != 0.0f || (font.Flags & 0x40) != 0)
               style |= TFStyle.fsItalic;
            // The force bold flag must be ignored if the font type is not Type1.
            if (font.FontType == TFontType.ftType1 && (font.Flags & 0x40000) != 0)
               style |= TFStyle.fsBold;
            BildFamilyNameAndStyle(font.FontName, ref familyName, ref style);
            if (m_PDFInst.SetFontEx(familyName, style, 10.0, true, DynaPDF.TCodepage.cpUnicode) < 0)
            {
               if ((font.Flags & 1) != 0) // Fixed pitch
               {
                  if (m_PDFInst.SetFont("Courier New", style, 10.0, true, DynaPDF.TCodepage.cpUnicode) < 0)
                     throw new Exception("-1");
               }else if (m_PDFInst.SetFont("Arial", style, 10.0, true, DynaPDF.TCodepage.cpUnicode) < 0)
                  throw new Exception("-1");
            }
            m_LastFont = m_Stack.IFont;
         }
      }
      private void Transform(ref TCTM M, ref double x, ref double y)
      {
         double ox = x;
         x = ox * M.a + y * M.c + M.x;
         y = ox * M.b + y * M.d + M.y;
      }
      private int WriteLine(int Index, ref TTextRec Record, String NewText)
      {
         // Process the remaining text in the kerning array
         ++Index;
         TTextRec rec;
         TTextRecordW src = new TTextRecordW();
         long ptr;
         String text;
         int i, j, p, count = 0;
         float spaceWidth = -m_Stack.SpaceWidth * 0.5f;
         while (Index < m_Records.Count())
         {
            rec = m_Records.GetItem(Index);
            if (rec.First != Record.First || rec.TmplHandle != Record.TmplHandle) break;
            ++Index;
            ++count;
         }
         /*
            We must always handle the text before and after the search text was found.
            Handling space characters correctly is the most important part if text
            should be replaced or deleted.
         */
         if (count > 0)
         {
            // The text was found multiple times in the text record
            ptr  = (long)m_Stack.Kerning;
            ptr += (Record.KernRecord * Marshal.SizeOf(src));
            CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
            // The old text was already deleted, write the new one
            WriteRecord(ref m_Matrix2, ref Record, ref src, NewText, NewText.Length);
            Index -= count;
            while (count-- > 0)
            {
               rec = m_Records.GetItem(Index);
               for (j = Record.KernRecord + 1; j < rec.KernRecord; j++)
               {
                  ptr += Marshal.SizeOf(src);
                  CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
                  AddKernSpace(src.Advance, spaceWidth, ref m_Matrix2);
                  text = Marshal.PtrToStringUni(src.Text, src.Length);
                  WriteText(ref m_Matrix2, text, text.Length);
               }
               ptr += Marshal.SizeOf(src);
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               if (rec.StrPos > -1 && m_SearchPos < m_SearchText.Length && m_SearchText[m_SearchPos] != ' ')
               {
                  m_Matrix4.x = -src.Advance;
                  m_Matrix2 = MulMatrix(ref m_Matrix2, ref m_Matrix4);
               }
               WriteRecord(ref m_Matrix2, ref rec, ref src, NewText, NewText.Length);
               Record = rec;
               ++Index;
            }
            // Handle the remaining records in the kerning array
            for (i = Record.KernRecord + 1; i < m_Stack.KerningCount; i++)
            {
               ptr += Marshal.SizeOf(src);
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               AddKernSpace(src.Advance, spaceWidth, ref m_Matrix2);
               text = Marshal.PtrToStringUni(src.Text, src.Length);
               WriteText(ref m_Matrix2, text, text.Length);
            }
         }else if (Record.KernRecord > -1)
         {
            ptr  = (long)m_Stack.Kerning;
            ptr += (Record.KernRecord * Marshal.SizeOf(src));
            CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
            WriteRecord(ref m_Matrix2, ref Record, ref src, NewText, NewText.Length);
            for (i = Record.KernRecord + 1; i < m_Stack.KerningCount; i++)
            {
               ptr += Marshal.SizeOf(src);
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               text = Marshal.PtrToStringUni(src.Text, src.Length);
               AddKernSpaceEx(src.Advance, spaceWidth, ref m_Matrix2);
               if (m_SearchPos < m_SearchText.Length)
               {
                  p = FindEndPattern(text, 0);
                  WriteTextEx(ref m_Matrix2, text, p);
               }else
                  WriteText(ref m_Matrix2, text, text.Length);
            }
         }else
         {
            // The text begins at the end of the previous text record.
            ptr = (long)m_Stack.Kerning;
            CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
            WriteRecord(ref m_Matrix3, ref Record, ref src, NewText, NewText.Length);
            for (i = 1; i < m_Stack.KerningCount; i++)
            {
               ptr += Marshal.SizeOf(src);
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               text = Marshal.PtrToStringUni(src.Text, src.Length);
               AddKernSpace(src.Advance, spaceWidth, ref m_Matrix3);
               WriteText(ref m_Matrix3, text, text.Length);
            }
            m_Matrix2 = m_Matrix3;
         }
         return WriteRemaining(Index, ref Record, NewText);
      }
      private void WriteRecord(ref TCTM M, ref TTextRec Record, ref TTextRecordW Source, String NewText, int Len)
      {
         int p;
         m_SearchPos = 0;
         String text = Marshal.PtrToStringUni(Source.Text, Source.Length);
         if (Record.StrPos < 0)
         {
            ++m_SearchPos;
            p = FindEndPattern(text, 0);
            // Write the new text if any
            if (Len > 0)
            {
               WriteText(ref M, NewText, NewText.Length);
            }
            if (p < Source.Length)
            {
               WriteTextEx(ref M, text, p);
            }
         }else
         {
            p = Record.StrPos;
            if (p > 0)
            {
               WriteText(ref M, text, p);
               // Write the new text if any
               if (Len > 0)
               {
                  WriteText(ref M, NewText, Len);
               }
               p = FindEndPattern(text, p);
               if (p < text.Length)
               {
                  WriteTextEx(ref M, text, p);
               }
            }else
            {
               p = FindEndPattern(text, 0);
               if (p < text.Length)
               {
                  WriteTextEx(ref M, text, p);
               }
               // Write the new text if any
               if (Len > 0)
               {
                  WriteText(ref M, NewText, NewText.Length);
               }
            }
         }
      }
      private int WriteRemaining(int Index, ref TTextRec Record, String NewText)
      {
         /*
            We handle here the rest of the text line if it is stored in separate text records.
            This is rather complex especially if the search text was found multiple times in the
            line and if the line consists in turn of multiple text records. We handle
            this case recursivley because this is much easier in comparison to a non-recursive
            version.
         */
         TCTM m1, m2 = new TCTM();
         TTextRec rec;
         TTextRecordW src = new TTextRecordW();
         float spaceWidth1;
         long ptr;
         String text;
         int nextIndex = -1;
         double x = 0.0, y = 0.0, spaceWidth2, distX = 0.0, distY = 0.0;
         int i, j, p, lastRecord, lastAlpha;
         if (!m_HaveMore)
         {
            ++m_RecordNumber;
            return Index;
         }
         if (Index >= m_Records.Count())
            lastRecord = 0x7FFFFFFF;
         else
         {
            rec        = m_Records.GetItem(Index);
            lastRecord = rec.First;
            if (!rec.NewLine) nextIndex = Index;
         }
         m_LastX = 0.0;
         m_LastY = 0.0;
         m_Matrix3 = m_Matrix2;
         // Get the end offset of the last text record. The current position of the
         // original text is different because the new text was already written to
         // the file.
         m_Matrix4.x = m_Stack.TextWidth;
         m_Matrix2   = MulMatrix(ref m_Stack.tm, ref m_Matrix4);
         m_Matrix2   = MulMatrix(ref m_Stack.ctm, ref m_Matrix2);
         Transform(ref m_Matrix2, ref m_LastX, ref m_LastY);
         CalcAlpha(ref m_Matrix2);
         lastAlpha   = m_Alpha;
         spaceWidth2 = m_Stack.SpaceWidth * GetScaleFactor(ref m_Matrix2) * 0.5;
         ++m_RecordNumber;
         // Was the search string found in multiple records?
         while (m_RecordNumber < Record.Last)
         {
            if ((m_HaveMore = m_PDFInst.GetPageText(ref m_Stack)) == false) return Index;
            ++m_RecordNumber;
            spaceWidth1 = -m_Stack.SpaceWidth / 2.0f;
            m_Matrix2   = MulMatrix(ref m_Stack.ctm, ref m_Stack.tm);
            x = 0.0;
            y = 0.0;
            Transform(ref m_Matrix2, ref x, ref y);
            AddSpace(x, y, spaceWidth2);
            m_PDFInst.ReplacePageTextA(null, ref m_Stack);
            ptr = (long)m_Stack.Kerning;
            for (i = 0; i < m_Stack.KerningCount; i++)
            {
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               text = Marshal.PtrToStringUni(src.Text, src.Length);
               AddKernSpaceEx(src.Advance, spaceWidth1, ref m_Matrix3);
               if (m_SearchPos < m_SearchText.Length)
               {
                  p = FindEndPattern(text, 0);
                  if (p < src.Length)
                  {
                     // We are now at the end of the already replaced text. Only the remaining text behind the
                     // search text must be written to the file.
                     WriteTextEx(ref m_Matrix3, text, p);
                  }
               }else
                  WriteText(ref m_Matrix3, text, text.Length);
               ptr += Marshal.SizeOf(src);
            }
            spaceWidth2 = m_Stack.SpaceWidth * GetScaleFactor(ref m_Matrix2) * 0.5;
            m_LastY     = 0.0;
            m_LastX     = m_Stack.TextWidth;
            Transform(ref m_Matrix2, ref m_LastX, ref m_LastY);
         }
         /*
            Now we come to the most complicated part: we must find all records which lie on the same
            text line and move the text behind the replaced one to avoid overlaping text or holes
            between text records. We must consider the spacing between records to preserve the original
            layout. Note that we deal with two text positions here: m_Matrix2 contains the coordinates
            of the original text while m_Matrix3 represents the position of the new text.
         */
         m1 = m_Matrix2;
         Invert(ref m1);
         CalcAlpha(ref m_Matrix2);
         lastAlpha = m_Alpha;
         while (m_HaveMore && m_RecordNumber < lastRecord)
         {
            if ((m_HaveMore = m_PDFInst.GetPageText(ref m_Stack)) == false) return Index;
            ++m_RecordNumber;
            spaceWidth1 = -m_Stack.SpaceWidth / 2.0f;
            CalcStrPos(ref m_Matrix2, ref x, ref y);
            if (m_Alpha != lastAlpha)
            {
               nextIndex = -1;
               break;
            }
            CalcDistance(ref m1, ref m_Matrix2, ref m2, ref distX, ref distY, x, y);
            if (distY > MAX_LINE_ERROR)
            {
               // We are on a new text line...
               nextIndex = -1;
               break;
            }else if (distX < -spaceWidth2 || distX > 6.0 * spaceWidth2)
            {
               // The distance is too large. We assume that the text should not be considered
               // as part of the current text line.
               nextIndex = -1;
               break;
            }
            m_PDFInst.ReplacePageTextA(null, ref m_Stack);     
            AddSpace(distX, spaceWidth2, ref m_Matrix3);
            ptr = (long)m_Stack.Kerning;
            for (i = 0; i < m_Stack.KerningCount; i++)
            {
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               text = Marshal.PtrToStringUni(src.Text, src.Length);
               AddKernSpaceEx(src.Advance, spaceWidth1, ref m_Matrix3);
               if (m_SearchPos < m_SearchText.Length)
               {
                  p = FindEndPattern(text, 0);
                  if (p < src.Length)
                  {
                     // We are now at the end of the already replaced text. Only the remaining text behind the
                     // search text must be written to the file.
                     WriteTextEx(ref m_Matrix3, text, p);
                  }
               }else
                  WriteText(ref m_Matrix3, text, text.Length);
               ptr += Marshal.SizeOf(src);
            }
            lastAlpha   = m_Alpha;
            spaceWidth2 = m_Stack.SpaceWidth * GetScaleFactor(ref m_Matrix2) * 0.5;
            m_LastY     = 0.0;
            m_LastX     = m_Stack.TextWidth;
            Transform(ref m_Matrix2, ref m_LastX, ref m_LastY);
         }
         if (nextIndex > -1)
         {
            rec = m_Records.GetItem(nextIndex);
            if (rec.NewLine) return Index;
            ++Index;
            m_HaveMore = m_PDFInst.GetPageText(ref m_Stack);
            // We are still in the same text line. We must check whether a space character must be deleted.
            // The rest of the line is now processed recursively.
            CalcStrPos(ref m_Matrix2, ref x, ref y);
            if (m_Alpha == 0)
               distX = x - m_LastX;
            else
               distX = CalcDistance(m_LastX, m_LastY, x, y);
            AddSpace(distX, spaceWidth2, ref m_Matrix3);
            if (rec.KernRecord > 0)
            {
               // Delete the string but preserve the kerning records before the string occurred.
               m_Stack.DeleteKerningAt = rec.KernRecord;
               m_PDFInst.ReplacePageTextA(null, ref m_Stack);
               m_Matrix4.x = 0.0;
               ptr = (long)m_Stack.Kerning;
               for (j = 0; j < rec.KernRecord; j++)
               {
                  CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
                  m_Matrix4.x -= src.Advance;
                  m_Matrix4.x += src.Width;
                  ptr += Marshal.SizeOf(src);
               }
               CPDF.CopyKernRecord(new IntPtr(ptr), ref src, Marshal.SizeOf(src));
               // If StrPos == -1 the first character is a space character which is emulated with kerning space.
               if (rec.StrPos > -1) m_Matrix4.x -= src.Advance;
               // Compute the string position
               m_Matrix2 = MulMatrix(ref m_Matrix3, ref m_Matrix4);
               Index = WriteLine(Index, ref rec, NewText);
            }else
            {
               m_PDFInst.ReplacePageTextA(null, ref m_Stack);
               m_Matrix2 = m_Matrix3;
               Index = WriteLine(Index, ref rec, NewText);
            }
         }
         return Index;
      }
      private void WriteText(ref TCTM M, String Text, int Len)
      {
         // Initialize the graphics state
         SetFont();
         // The new text is written in red color so that you can better find it
         m_PDFInst.SetColorSpace(TPDFColorSpace.csDeviceRGB);
         m_PDFInst.SetFillColor(CPDF.PDF_RED);
         m_PDFInst.SetStrokeColor(CPDF.PDF_RED);
         m_PDFInst.SetTextDrawMode(m_Stack.DrawMode);
/*
         // This is the normal initalization
         m_PDFInst.SetFillColorSpace(m_Stack.FillCS);
         m_PDFInst.SetFillColor(m_Stack.FillColor);
         m_PDFInst.SetStrokeColorSpace(m_Stack.StrokeCS);
         m_PDFInst.SetStrokeColor(m_Stack.StrokeColor);
         m_PDFInst.SetTextDrawMode(m_Stack.DrawMode);
*/
         m_PDFInst.ChangeFontSize(m_Stack.FontSize);
         m_PDFInst.WriteTextMatrixExW(ref M, Text, Len);
         // Compute the end offset of the text
         m_Matrix4.x = m_PDFInst.GetTextWidthExW(Text, Len);
         M = MulMatrix(ref M, ref m_Matrix4);
      }
      private void WriteTextEx(ref TCTM M, String Text, int First)
      {
         // Initialize the graphics state
         SetFont();
         // The new text is written in red color so that you can better find it
         m_PDFInst.SetColorSpace(TPDFColorSpace.csDeviceRGB);
         m_PDFInst.SetFillColor(CPDF.PDF_RED);
         m_PDFInst.SetStrokeColor(CPDF.PDF_RED);
         m_PDFInst.SetTextDrawMode(m_Stack.DrawMode);
/*
         // This is the normal initalization
         m_PDFInst.SetFillColorSpace(m_Stack.FillCS);
         m_PDFInst.SetFillColor(m_Stack.FillColor);
         m_PDFInst.SetStrokeColorSpace(m_Stack.StrokeCS);
         m_PDFInst.SetStrokeColor(m_Stack.StrokeColor);
         m_PDFInst.SetTextDrawMode(m_Stack.DrawMode);
*/
         m_PDFInst.ChangeFontSize(m_Stack.FontSize);
         // Get the substring
         String txt = Text.Substring(First);
         m_PDFInst.WriteTextMatrixExW(ref M, txt, txt.Length);
         // Compute the end offset of the text
         m_Matrix4.x = m_PDFInst.GetTextWidthExW(txt, txt.Length);
         M = MulMatrix(ref M, ref m_Matrix4);
      }

      private int          m_Alpha;
      private int          m_CurrTmpl;
      private int          m_First;
      private bool         m_HaveMore;
      private int          m_KernRecord;
      private IntPtr       m_LastFont;
      private double       m_LastX;
      private double       m_LastY;
      private TCTM         m_Matrix1;
      private TCTM         m_Matrix2;
      private TCTM         m_Matrix3;
      private TCTM         m_Matrix4;
      private bool         m_NewLine;
      private CPDF         m_PDFInst;
      private int          m_RecordNumber;
      private CTextRecords m_Records;
      private int          m_SearchPos;
      private String       m_SearchText;
      private TPDFStack    m_Stack;
      private int          m_StrPos;
      private CIntList     m_Templates;
	}
}
